@@ -4,8 +4,8 @@ from pyjjasim import *
 import matplotlib.pyplot as plt
 
 # If animation does not start; try the following lines of code:
-# import matplotlib
-# matplotlib.use("TkAgg")
+import matplotlib
+matplotlib.use("TkAgg")
 
 """
 Static Example 0: Single vortex
@@ -21,7 +21,7 @@ if __name__ == "__main__":
     dt = 0.05
     Nt = 1000
     f = 0.2
-    Is = 0.5 * sq_array.horizontal_junctions()
+    Is = 0.5 * sq_array.current_base(angle=0)
     T = 0.02
     problem = TimeEvolutionProblem(sq_array, time_step=dt,
                                    time_step_count=Nt, current_sources=Is,
